@@ -3,7 +3,7 @@
  *  BlueZ - Bluetooth protocol stack for Linux
  *
  *  Copyright (C) 2006-2007  Nokia Corporation
- *  Copyright (C) 2004-2008  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2004-2009  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -656,7 +656,8 @@ int telephony_event_reporting_rsp(void *telephony_device, cme_error_t err)
 	if (hs->state != HEADSET_STATE_CONNECT_IN_PROGRESS)
 		return 0;
 
-	if (ag.features & AG_FEATURE_THREE_WAY_CALLING)
+	if (hs->hf_features & HF_FEATURE_CALL_WAITING_AND_3WAY &&
+			ag.features & AG_FEATURE_THREE_WAY_CALLING)
 		return 0;
 
 	hfp_slc_complete(device);
@@ -990,7 +991,7 @@ static int call_waiting_notify(struct audio_device *device, const char *buf)
 
 	if (buf[8] == '1') {
 		hs->cwa_enabled = TRUE;
-		debug("Call waiting notifiaction enabled for headset %p", hs);
+		debug("Call waiting notification enabled for headset %p", hs);
 	} else {
 		hs->cwa_enabled = FALSE;
 		debug("Call waiting notification disabled for headset %p", hs);
@@ -1122,11 +1123,8 @@ static gboolean rfcomm_io_cb(GIOChannel *chan, GIOCondition cond,
 {
 	struct headset *hs;
 	unsigned char buf[BUF_SIZE];
-	char *cr;
 	gsize bytes_read = 0;
 	gsize free_space;
-	int err;
-	off_t cmd_len;
 
 	if (cond & G_IO_NVAL)
 		return FALSE;
@@ -1156,27 +1154,39 @@ static gboolean rfcomm_io_cb(GIOChannel *chan, GIOCondition cond,
 	 * functions */
 	hs->buf[hs->data_start + hs->data_length] = '\0';
 
-	cr = strchr(&hs->buf[hs->data_start], '\r');
-	if (!cr)
-		return TRUE;
+	while (hs->data_length > 0) {
+		char *cr;
+		int err;
+		off_t cmd_len;
 
-	cmd_len = 1 + (off_t) cr - (off_t) &hs->buf[hs->data_start];
-	*cr = '\0';
+		cr = strchr(&hs->buf[hs->data_start], '\r');
+		if (!cr)
+			break;
 
-	err = handle_event(device, &hs->buf[hs->data_start]);
-	if (err == -EINVAL) {
-		error("Badly formated or unrecognized command: %s",
-				&hs->buf[hs->data_start]);
-		err = headset_send(hs, "\r\nERROR\r\n");
-	} else if (err < 0)
-		error("Error handling command %s: %s (%d)",
-			&hs->buf[hs->data_start], strerror(-err), -err);
+		cmd_len = 1 + (off_t) cr - (off_t) &hs->buf[hs->data_start];
+		*cr = '\0';
 
-	hs->data_start += cmd_len;
-	hs->data_length -= cmd_len;
+		if (cmd_len > 1)
+			err = handle_event(device, &hs->buf[hs->data_start]);
+		else
+			/* Silently skip empty commands */
+			err = 0;
 
-	if (!hs->data_length)
-		hs->data_start = 0;
+		if (err == -EINVAL) {
+			error("Badly formated or unrecognized command: %s",
+					&hs->buf[hs->data_start]);
+			err = headset_send(hs, "\r\nERROR\r\n");
+		} else if (err < 0)
+			error("Error handling command %s: %s (%d)",
+						&hs->buf[hs->data_start],
+						strerror(-err), -err);
+
+		hs->data_start += cmd_len;
+		hs->data_length -= cmd_len;
+
+		if (!hs->data_length)
+			hs->data_start = 0;
+	}
 
 	return TRUE;
 
