@@ -3,7 +3,7 @@
  *  BlueZ - Bluetooth protocol stack for Linux
  *
  *  Copyright (C) 2006-2007  Nokia Corporation
- *  Copyright (C) 2004-2008  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2004-2009  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -314,6 +314,13 @@ static void headset_resume_complete(struct audio_device *dev, void *user_data)
 		goto failed;
 	}
 
+	client->data_fd = headset_get_sco_fd(dev);
+	if (client->data_fd < 0) {
+		error("Unable to get a SCO fd");
+		headset_unlock(dev, hs->lock);
+		goto failed;
+	}
+
 	memset(buf, 0, sizeof(buf));
 	rsp->h.type = BT_RESPONSE;
 	rsp->h.name = BT_START_STREAM;
@@ -328,8 +335,6 @@ static void headset_resume_complete(struct audio_device *dev, void *user_data)
 
 	unix_ipc_sendmsg(client, &ind->h);
 
-	client->data_fd = headset_get_sco_fd(dev);
-
 	if (unix_sendmsg_fd(client->sock, client->data_fd) < 0) {
 		error("unix_sendmsg_fd: %s(%d)", strerror(errno), errno);
 		headset_unlock(client->dev, hs->lock);
@@ -643,8 +648,6 @@ static void start_discovery(struct audio_device *dev, struct unix_client *client
 	struct a2dp_data *a2dp;
 	int err = 0;
 
-	client->type = select_service(dev, client->interface);
-
 	switch (client->type) {
 	case TYPE_SINK:
 		a2dp = &client->d.a2dp;
@@ -686,8 +689,6 @@ static void start_config(struct audio_device *dev, struct unix_client *client)
 	struct headset_data *hs;
 	unsigned int id;
 
-	client->type = select_service(dev, client->interface);
-
 	switch (client->type) {
 	case TYPE_SINK:
 		a2dp = &client->d.a2dp;
@@ -753,8 +754,6 @@ static void start_resume(struct audio_device *dev, struct unix_client *client)
 	struct headset_data *hs;
 	unsigned int id;
 
-	client->type = select_service(dev, client->interface);
-
 	switch (client->type) {
 	case TYPE_SINK:
 		a2dp = &client->d.a2dp;
@@ -808,8 +807,6 @@ static void start_suspend(struct audio_device *dev, struct unix_client *client)
 	struct headset_data *hs;
 	unsigned int id;
 
-	client->type = select_service(dev, client->interface);
-
 	switch (client->type) {
 	case TYPE_SINK:
 		a2dp = &client->d.a2dp;
@@ -865,8 +862,8 @@ static void handle_getcapabilities_req(struct unix_client *client,
 	str2ba(req->device, &bdaddr);
 
 	if (client->interface) {
-		g_free(client->interface);
-		client->interface = NULL;
+		error("Got GET_CAPABILITIES for an initialized client");
+		goto failed;
 	}
 
 	if (req->transport == BT_CAPABILITIES_TRANSPORT_SCO)
@@ -878,15 +875,19 @@ static void handle_getcapabilities_req(struct unix_client *client,
 		goto failed;
 
 	dev = manager_find_device(&bdaddr, client->interface, TRUE);
+	if (!dev && (req->flags & BT_FLAG_AUTOCONNECT))
+		dev = manager_find_device(&bdaddr, client->interface, FALSE);
+
 	if (!dev) {
-		if (req->flags & BT_FLAG_AUTOCONNECT)
-			dev = manager_find_device(&bdaddr, client->interface, FALSE);
-		else
-			goto failed;
+		error("Unable to find a matching device");
+		goto failed;
 	}
 
-	if (!dev)
+	client->type = select_service(dev, client->interface);
+	if (client->type == TYPE_NONE) {
+		error("No matching service found");
 		goto failed;
+	}
 
 	start_discovery(dev, client);
 
@@ -899,7 +900,10 @@ failed:
 static int handle_sco_transport(struct unix_client *client,
 				struct bt_set_configuration_req *req)
 {
-	client->interface = g_strdup(AUDIO_HEADSET_INTERFACE);
+	if (!client->interface)
+		client->interface = g_strdup(AUDIO_HEADSET_INTERFACE);
+	else if (!g_str_equal(client->interface, AUDIO_HEADSET_INTERFACE))
+		return -EIO;
 
 	debug("config sco - device = %s access_mode = %u", req->device,
 			req->access_mode);
@@ -914,11 +918,15 @@ static int handle_a2dp_transport(struct unix_client *client,
 	struct sbc_codec_cap sbc_cap;
 	struct mpeg_codec_cap mpeg_cap;
 
-	client->interface = g_strdup(AUDIO_SINK_INTERFACE);
+	if (!client->interface)
+		client->interface = g_strdup(AUDIO_SINK_INTERFACE);
+	else if (!g_str_equal(client->interface, AUDIO_SINK_INTERFACE))
+		return -EIO;
 
 	if (client->caps) {
 		g_slist_foreach(client->caps, (GFunc) g_free, NULL);
 		g_slist_free(client->caps);
+		client->caps = NULL;
 	}
 
 	media_transport = avdtp_service_cap_new(AVDTP_MEDIA_TRANSPORT,
@@ -997,11 +1005,6 @@ static void handle_setconfiguration_req(struct unix_client *client,
 
 	str2ba(req->device, &bdaddr);
 
-	if (client->interface) {
-		g_free(client->interface);
-		client->interface = NULL;
-	}
-
 	if (req->codec.transport == BT_CAPABILITIES_TRANSPORT_SCO) {
 		err = handle_sco_transport(client, req);
 		if (err < 0) {
@@ -1117,7 +1120,7 @@ static gboolean client_cb(GIOChannel *chan, GIOCondition cond, gpointer data)
 
 	memset(buf, 0, sizeof(buf));
 
-	len = recv(client->sock, buf, sizeof(buf), MSG_WAITALL);
+	len = recv(client->sock, buf, sizeof(buf), 0);
 	if (len < 0) {
 		error("recv: %s (%d)", strerror(errno), errno);
 		goto failed;
@@ -1195,6 +1198,7 @@ static gboolean server_cb(GIOChannel *chan, GIOCondition cond, gpointer data)
 	}
 
 	debug("Accepted new client connection on unix socket (fd=%d)", cli_sk);
+	set_nonblocking(cli_sk);
 
 	client = g_new0(struct unix_client, 1);
 	client->sock = cli_sk;
