@@ -61,6 +61,34 @@
 
 static DBusConnection *connection = NULL;
 
+static gboolean get_adapter_and_device(bdaddr_t *src, bdaddr_t *dst,
+					struct btd_adapter **adapter,
+					struct btd_device **device,
+					gboolean create)
+{
+	char peer_addr[18];
+
+	*adapter = manager_find_adapter(src);
+	if (!*adapter) {
+		error("Unable to find matching adapter");
+		return FALSE;
+	}
+
+	ba2str(dst, peer_addr);
+
+	if (create)
+		*device = adapter_get_device(connection, *adapter, peer_addr);
+	else
+		*device = adapter_find_device(*adapter, peer_addr);
+
+	if (create && !*device) {
+		error("Unable to get device object!");
+		return FALSE;
+	}
+
+	return TRUE;
+}
+
 const char *class_to_icon(uint32_t class)
 {
 	switch ((class & 0x1f00) >> 8) {
@@ -86,6 +114,8 @@ const char *class_to_icon(uint32_t class)
 			return "audio-card";	/* Headset */
 		case 0x06:
 			return "audio-card";	/* Headphone */
+		default:
+			return "audio-card";	/* Other audio device */
 		}
 		break;
 	case 0x05:
@@ -160,7 +190,8 @@ static void pincode_cb(struct agent *agent, DBusError *err, const char *pincode,
 	bacpy(&pr.bdaddr, &dba);
 	memcpy(pr.pin_code, pincode, len);
 	pr.pin_len = len;
-	hci_send_cmd(dev, OGF_LINK_CTL, OCF_PIN_CODE_REPLY, PIN_CODE_REPLY_CP_SIZE, &pr);
+	hci_send_cmd(dev, OGF_LINK_CTL, OCF_PIN_CODE_REPLY,
+						PIN_CODE_REPLY_CP_SIZE, &pr);
 
 done:
 	hci_close_dev(dev);
@@ -168,31 +199,19 @@ done:
 
 int hcid_dbus_request_pin(int dev, bdaddr_t *sba, struct hci_conn_info *ci)
 {
-	char addr[18];
 	struct btd_adapter *adapter;
 	struct btd_device *device;
 
-	adapter = manager_find_adapter(sba);
-	if (!adapter) {
-		error("No matching adapter found");
-		return -1;
-	}
-
-	ba2str(&ci->bdaddr, addr);
-
-	device = adapter_find_device(adapter, addr);
-	/* Check if the adapter is not pairable and if there isn't a bonding in
-	 * progress */
-	if (!adapter_is_pairable(adapter) &&
-			!(device && device_is_bonding(device, NULL)))
-		return -EPERM;
-
-	device = adapter_get_device(connection, adapter, addr);
-	if (!device)
+	if (!get_adapter_and_device(sba, &ci->bdaddr, &adapter, &device, TRUE))
 		return -ENODEV;
 
-	return device_request_authentication(device, AUTH_TYPE_PINCODE,
-					0, pincode_cb);
+	/* Check if the adapter is not pairable and if there isn't a bonding in
+	 * progress */
+	if (!adapter_is_pairable(adapter) && !device_is_bonding(device, NULL))
+		return -EPERM;
+
+	return device_request_authentication(device, AUTH_TYPE_PINCODE, 0,
+								pincode_cb);
 }
 
 static void confirm_cb(struct agent *agent, DBusError *err, void *user_data)
@@ -294,15 +313,11 @@ int hcid_dbus_user_confirm(bdaddr_t *sba, bdaddr_t *dba, uint32_t passkey)
 {
 	struct btd_adapter *adapter;
 	struct btd_device *device;
-	char addr[18];
 	uint8_t remcap, remauth, type;
 	uint16_t dev_id;
 
-	adapter = manager_find_adapter(sba);
-	if (!adapter) {
-		error("No matching adapter found");
-		return -1;
-	}
+	if (!get_adapter_and_device(sba, dba, &adapter, &device, TRUE))
+		return -ENODEV;
 
 	dev_id = adapter_get_dev_id(adapter);
 
@@ -325,14 +340,6 @@ int hcid_dbus_user_confirm(bdaddr_t *sba, bdaddr_t *dba, uint32_t passkey)
 
 	debug("confirm authentication requirement is 0x%02x", type);
 
-	ba2str(dba, addr);
-
-	device = adapter_get_device(connection, adapter, addr);
-	if (!device) {
-		error("Device creation failed");
-		return -1;
-	}
-
 	remcap = device_get_cap(device);
 	remauth = device_get_auth(device);
 
@@ -340,7 +347,8 @@ int hcid_dbus_user_confirm(bdaddr_t *sba, bdaddr_t *dba, uint32_t passkey)
 	debug("remote authentication requirement is 0x%02x", remauth);
 
 	/* If no side requires MITM protection; auto-accept */
-	if (!(remauth & 0x01) && (!(type & 0x01) || remcap == 0x03)) {
+	if (!(remauth & 0x01) &&
+			(type == 0xff || !(type & 0x01) || remcap == 0x03)) {
 		int dd;
 
 		dd = hci_open_dev(dev_id);
@@ -368,69 +376,36 @@ int hcid_dbus_user_passkey(bdaddr_t *sba, bdaddr_t *dba)
 {
 	struct btd_adapter *adapter;
 	struct btd_device *device;
-	char addr[18];
 
-	adapter = manager_find_adapter(sba);
-	if (!adapter) {
-		error("No matching adapter found");
-		return -1;
-	}
-
-	ba2str(dba, addr);
-
-	device = adapter_get_device(connection, adapter, addr);
-
-	if (!device)
+	if (!get_adapter_and_device(sba, dba, &adapter, &device, TRUE))
 		return -ENODEV;
 
-	return device_request_authentication(device, AUTH_TYPE_PASSKEY,
-					0, passkey_cb);
+	return device_request_authentication(device, AUTH_TYPE_PASSKEY, 0,
+								passkey_cb);
 }
 
 int hcid_dbus_user_notify(bdaddr_t *sba, bdaddr_t *dba, uint32_t passkey)
 {
 	struct btd_adapter *adapter;
 	struct btd_device *device;
-	char addr[18];
 
-	adapter = manager_find_adapter(sba);
-	if (!adapter) {
-		error("No matching adapter found");
-		return -1;
-	}
-
-	ba2str(dba, addr);
-
-	device = adapter_get_device(connection, adapter, addr);
-	if (!device)
+	if (!get_adapter_and_device(sba, dba, &adapter, &device, TRUE))
 		return -ENODEV;
 
 	return device_request_authentication(device, AUTH_TYPE_NOTIFY,
-					passkey, NULL);
+								passkey, NULL);
 }
 
 void hcid_dbus_bonding_process_complete(bdaddr_t *local, bdaddr_t *peer,
-					uint8_t status)
+								uint8_t status)
 {
 	struct btd_adapter *adapter;
-	char peer_addr[18];
 	struct btd_device *device;
 
 	debug("hcid_dbus_bonding_process_complete: status=%02x", status);
 
-	adapter = manager_find_adapter(local);
-	if (!adapter) {
-		error("Unable to find matching adapter");
+	if (!get_adapter_and_device(local, peer, &adapter, &device, TRUE))
 		return;
-	}
-
-	ba2str(peer, peer_addr);
-
-	device = adapter_find_device(adapter, peer_addr);
-	if (!device) {
-		error("Unable to get device object!");
-		return;
-	}
 
 	if (!device_is_authenticating(device)) {
 		/* This means that there was no pending PIN or SSP token
@@ -445,6 +420,20 @@ void hcid_dbus_bonding_process_complete(bdaddr_t *local, bdaddr_t *peer,
 	device_bonding_complete(device, status);
 }
 
+void hcid_dbus_simple_pairing_complete(bdaddr_t *local, bdaddr_t *peer,
+					uint8_t status)
+{
+	struct btd_adapter *adapter;
+	struct btd_device *device;
+
+	debug("hcid_dbus_simple_pairing_complete: status=%02x", status);
+
+	if (!get_adapter_and_device(local, peer, &adapter, &device, TRUE))
+		return;
+
+	device_simple_pairing_complete(device, status);
+}
+
 void hcid_dbus_inquiry_start(bdaddr_t *local)
 {
 	struct btd_adapter *adapter;
@@ -515,7 +504,7 @@ static int found_device_req_name(struct btd_adapter *adapter)
 		cp.pscan_rep_mode = 0x02;
 
 		if (hci_send_req(dd, &rq, HCI_REQ_TIMEOUT) < 0)
-			error("Unable to send the HCI remote name request: %s (%d)",
+			error("Unable to send HCI remote name req: %s (%d)",
 						strerror(errno), errno);
 
 		if (!rp.status) {
@@ -808,8 +797,6 @@ void hcid_dbus_inquiry_result(bdaddr_t *local, bdaddr_t *peer, uint32_t class,
 
 void hcid_dbus_remote_class(bdaddr_t *local, bdaddr_t *peer, uint32_t class)
 {
-	char peer_addr[18];
-	const char *paddr = peer_addr;
 	uint32_t old_class = 0;
 	struct btd_adapter *adapter;
 	struct btd_device *device;
@@ -820,14 +807,8 @@ void hcid_dbus_remote_class(bdaddr_t *local, bdaddr_t *peer, uint32_t class)
 	if (old_class == class)
 		return;
 
-	adapter = manager_find_adapter(local);
-	if (!adapter) {
-		error("No matching adapter found");
+	if (!get_adapter_and_device(local, peer, &adapter, &device, FALSE))
 		return;
-	}
-
-	ba2str(peer, peer_addr);
-	device = adapter_find_device(adapter, paddr);
 
 	if (!device)
 		return;
@@ -847,11 +828,8 @@ void hcid_dbus_remote_name(bdaddr_t *local, bdaddr_t *peer, uint8_t status,
 	struct btd_device *device;
 	struct remote_dev_info match, *dev_info;
 
-	adapter = manager_find_adapter(local);
-	if (!adapter) {
-		error("No matching adapter found");
+	if (!get_adapter_and_device(local, peer, &adapter, &device, FALSE))
 		return;
-	}
 
 	ba2str(local, srcaddr);
 	ba2str(peer, dstaddr);
@@ -891,20 +869,19 @@ void hcid_dbus_remote_name(bdaddr_t *local, bdaddr_t *peer, uint8_t status,
 				NULL);
 	}
 
-	device = adapter_find_device(adapter, dstaddr);
 	if (device) {
 		char alias[248];
 		const char *dev_path = device_get_path(device);
 
 		emit_property_changed(connection, dev_path,
-				DEVICE_INTERFACE, "Name",
-				DBUS_TYPE_STRING, &name);
+						DEVICE_INTERFACE, "Name",
+						DBUS_TYPE_STRING, &name);
 
 		if (read_device_alias(srcaddr, dstaddr,
 					alias, sizeof(alias)) < 1)
 			emit_property_changed(connection, dev_path,
-					DEVICE_INTERFACE, "Alias",
-					DBUS_TYPE_STRING, &name);
+						DEVICE_INTERFACE, "Alias",
+						DBUS_TYPE_STRING, &name);
 	}
 
 proceed:
@@ -921,65 +898,83 @@ proceed:
 	adapter_set_state(adapter, state);
 }
 
-void hcid_dbus_link_key_notify(bdaddr_t *local, bdaddr_t *peer,
-				uint8_t key_type, uint8_t old_key_type)
+int hcid_dbus_link_key_notify(bdaddr_t *local, bdaddr_t *peer,
+				uint8_t *key, uint8_t key_type,
+				int pin_length, uint8_t old_key_type)
 {
-	char peer_addr[18];
 	struct btd_device *device;
 	struct btd_adapter *adapter;
+	uint8_t local_auth = 0xff, remote_auth, new_key_type;
 
-	adapter = manager_find_adapter(local);
-	if (!adapter) {
-		error("No matching adapter found");
-		return;
-	}
+	if (!get_adapter_and_device(local, peer, &adapter, &device, TRUE))
+		return -ENODEV;
 
-	ba2str(peer, peer_addr);
+	if (key_type == 0x06 && old_key_type != 0xff)
+		new_key_type = old_key_type;
+	else
+		new_key_type = key_type;
 
-	device = adapter_get_device(connection, adapter, peer_addr);
-	if (!device) {
-		error("Couldn't get a device object for %s", peer_addr);
-		return;
+	get_auth_requirements(local, peer, &local_auth);
+	remote_auth = device_get_auth(device);
+
+	debug("local auth 0x%02x and remote auth 0x%02x",
+					local_auth, remote_auth);
+
+	/* Only store the link key if one of the following is true:
+	 * 1. this is a legacy link key
+	 * 2. this is a changed combination key and there was a previously
+	 *    stored one
+	 * 3. neither local nor remote side had no-bonding as a requirement
+	 * 4. the local side had dedicated bonding as a requirement
+	 * 5. the remote side is using dedicated bonding since in that case
+	 *    also the local requirements are set to dedicated bonding
+	 */
+	if (key_type < 0x03 || (key_type == 0x06 && old_key_type != 0xff) ||
+				(local_auth > 0x01 && remote_auth > 0x01) ||
+				(local_auth == 0x02 || local_auth == 0x03) ||
+				(remote_auth == 0x02 || remote_auth == 0x03)) {
+		int err;
+
+		debug("storing link key of type 0x%02x", key_type);
+
+		err = write_link_key(local, peer, key, new_key_type,
+								pin_length);
+		if (err < 0) {
+			error("write_link_key: %s (%d)", strerror(-err), -err);
+			return err;
+		}
+
+		device_set_temporary(device, FALSE);
 	}
 
 	/* If this is not the first link key set a flag so a subsequent auth
 	 * complete event doesn't trigger SDP */
-	if (old_key_type != 0xFF)
+	if (old_key_type != 0xff)
 		device_set_renewed_key(device, TRUE);
 
-	device_set_temporary(device, FALSE);
-
 	if (!device_is_connected(device))
 		device_set_secmode3_conn(device, TRUE);
-	else if (!device_is_bonding(device, NULL) && old_key_type == 0xFF)
+	else if (!device_is_bonding(device, NULL) && old_key_type == 0xff)
 		hcid_dbus_bonding_process_complete(local, peer, 0);
+
+	return 0;
 }
 
 void hcid_dbus_conn_complete(bdaddr_t *local, uint8_t status, uint16_t handle,
 				bdaddr_t *peer)
 {
-	char peer_addr[18];
 	struct btd_adapter *adapter;
 	struct btd_device *device;
 
-	adapter = manager_find_adapter(local);
-	if (!adapter) {
-		error("No matching adapter found");
+	if (!get_adapter_and_device(local, peer, &adapter, &device, TRUE))
 		return;
-	}
-
-	ba2str(peer, peer_addr);
-
-	device = adapter_get_device(connection, adapter, peer_addr);
-	if (!device) {
-		error("No matching device found");
-		return;
-	}
 
 	if (status) {
 		device_set_secmode3_conn(device, FALSE);
 		if (device_is_bonding(device, NULL))
 			device_bonding_complete(device, status);
+		if (device_is_temporary(device))
+			adapter_remove_device(connection, adapter, device);
 		return;
 	}
 
@@ -1054,20 +1049,20 @@ int set_major_and_minor_class(int dd, const uint8_t *cls,
 
 void hcid_dbus_setname_complete(bdaddr_t *local)
 {
+	struct btd_adapter *adapter;
 	int id, dd = -1;
 	read_local_name_rp rp;
 	struct hci_request rq;
 	const char *pname = (char *) rp.name;
-	char local_addr[18], name[249];
+	char name[249];
 
-	ba2str(local, local_addr);
-
-	id = hci_devid(local_addr);
-	if (id < 0) {
-		error("No matching device id for %s", local_addr);
+	adapter = manager_find_adapter(local);
+	if (!adapter) {
+		error("No matching adapter found");
 		return;
 	}
 
+	id = adapter_get_dev_id(adapter);
 	dd = hci_open_dev(id);
 	if (dd < 0) {
 		error("HCI device open failed: hci%d", id);
@@ -1095,6 +1090,8 @@ void hcid_dbus_setname_complete(bdaddr_t *local)
 	strncpy(name, pname, sizeof(name) - 1);
 	name[248] = '\0';
 	pname = name;
+
+	adapter_name_changed(adapter, pname);
 }
 
 void hcid_dbus_setscan_enable_complete(bdaddr_t *local)
@@ -1204,7 +1201,7 @@ void hcid_dbus_write_simple_pairing_mode_complete(bdaddr_t *local)
 
 	if (hci_read_simple_pairing_mode(dd, &mode,
 						HCI_REQ_TIMEOUT) < 0) {
-		error("Can't read class of adapter on %s: %s(%d)",
+		error("Can't read simple pairing mode for %s: %s(%d)",
 					path, strerror(errno), errno);
 		hci_close_dev(dd);
 		return;
@@ -1221,24 +1218,17 @@ int hcid_dbus_get_io_cap(bdaddr_t *local, bdaddr_t *remote,
 	struct btd_adapter *adapter;
 	struct btd_device *device;
 	struct agent *agent = NULL;
-	char addr[18];
 
-	adapter = manager_find_adapter(local);
-	if (!adapter) {
-		error("No matching adapter found");
-		return -1;
-	}
+	if (!get_adapter_and_device(local, remote, &adapter, &device, TRUE))
+		return -ENODEV;
 
 	if (get_auth_requirements(local, remote, auth) < 0)
 		return -1;
 
 	debug("initial authentication requirement is 0x%02x", *auth);
 
-	ba2str(remote, addr);
-
-	device = adapter_find_device(adapter, addr);
-	if (!device)
-		return -ENODEV;
+	if (*auth == 0xff)
+		*auth = device_get_auth(device);
 
 	/* Check if the adapter is not pairable and if there isn't a bonding
 	 * in progress */
@@ -1293,7 +1283,8 @@ int hcid_dbus_get_io_cap(bdaddr_t *local, bdaddr_t *remote,
 		}
 
 		/* If remote requires MITM then also require it */
-		if (device_get_auth(device) & 0x01)
+		if (device_get_auth(device) != 0xff &&
+					(device_get_auth(device) & 0x01))
 			*auth |= 0x01;
 	}
 
@@ -1310,21 +1301,12 @@ int hcid_dbus_set_io_cap(bdaddr_t *local, bdaddr_t *remote,
 {
 	struct btd_adapter *adapter;
 	struct btd_device *device;
-	char addr[18];
 
-	adapter = manager_find_adapter(local);
-	if (!adapter) {
-		error("No matching adapter found");
-		return -1;
-	}
+	if (!get_adapter_and_device(local, remote, &adapter, &device, TRUE))
+		return -ENODEV;
 
-	ba2str(remote, addr);
-
-	device = adapter_get_device(connection, adapter, addr);
-	if (device) {
-		device_set_cap(device, cap);
-		device_set_auth(device, auth);
-	}
+	device_set_cap(device, cap);
+	device_set_auth(device, auth);
 
 	return 0;
 }
@@ -1402,7 +1384,8 @@ int cancel_discovery(struct btd_adapter *adapter)
 
 	dev = adapter_search_found_devices(adapter, &match);
 	if (dev) {
-		if (remote_name_cancel(dd, &dev->bdaddr, HCI_REQ_TIMEOUT) < 0) {
+		if (remote_name_cancel(dd, &dev->bdaddr,
+							HCI_REQ_TIMEOUT) < 0) {
 			error("Read remote name cancel failed: %s, (%d)",
 					strerror(errno), errno);
 			err = -errno;
@@ -1460,7 +1443,8 @@ int cancel_periodic_discovery(struct btd_adapter *adapter)
 
 	dev = adapter_search_found_devices(adapter, &match);
 	if (dev) {
-		if (remote_name_cancel(dd, &dev->bdaddr, HCI_REQ_TIMEOUT) < 0) {
+		if (remote_name_cancel(dd, &dev->bdaddr,
+							HCI_REQ_TIMEOUT) < 0) {
 			error("Read remote name cancel failed: %s, (%d)",
 					strerror(errno), errno);
 			err = -errno;
