@@ -94,7 +94,8 @@ struct btd_device {
 	struct agent	*agent;
 	guint		disconn_timer;
 	int		discov_active;		/* Service discovery active */
-	char		*discov_requestor;	/* discovery requestor unique name */
+	char		*discov_requestor;	/* discovery requestor unique
+						 * name */
 	guint		discov_listener;
 	guint		discov_timer;
 	struct bonding_req *bonding;
@@ -346,7 +347,6 @@ static DBusMessage *set_trust(DBusConnection *conn, DBusMessage *msg,
 	char srcaddr[18], dstaddr[18];
 	bdaddr_t src;
 
-
 	adapter_get_address(adapter, &src);
 	ba2str(&src, srcaddr);
 	ba2str(&device->bdaddr, dstaddr);
@@ -430,7 +430,7 @@ static DBusMessage *discover_services(DBusConnection *conn,
 
 	if (device->discov_active)
 		return g_dbus_create_error(msg, ERROR_INTERFACE ".InProgress",
-							"Discover in progress");
+						"Discover in progress");
 
 	if (dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &pattern,
 						DBUS_TYPE_INVALID) == FALSE)
@@ -465,6 +465,7 @@ static DBusMessage *cancel_discover(DBusConnection *conn,
 {
 	struct btd_device *device = user_data;
 	struct btd_adapter *adapter = device->adapter;
+	const char *sender = dbus_message_get_sender(msg);
 	bdaddr_t src;
 
 	adapter_get_address(adapter, &src);
@@ -476,7 +477,7 @@ static DBusMessage *cancel_discover(DBusConnection *conn,
 
 	/* only the discover requestor can cancel the inquiry process */
 	if (!device->discov_requestor ||
-			strcmp(device->discov_requestor, dbus_message_get_sender(msg)))
+				!g_str_equal(device->discov_requestor, sender))
 		return g_dbus_create_error(msg,
 				ERROR_INTERFACE ".NotAuthorized",
 				"Not Authorized");
@@ -515,8 +516,8 @@ fail:
 	return FALSE;
 }
 
-static DBusMessage *disconnect(DBusConnection *conn,
-					DBusMessage *msg, void *user_data)
+static DBusMessage *disconnect(DBusConnection *conn, DBusMessage *msg,
+							void *user_data)
 {
 	struct btd_device *device = user_data;
 
@@ -556,8 +557,9 @@ gboolean device_is_connected(struct btd_device *device)
 	return (device->handle != 0);
 }
 
-static void device_set_connected(struct btd_device *device, DBusConnection *conn,
-			gboolean connected)
+static void device_set_connected(struct btd_device *device,
+					DBusConnection *conn,
+					gboolean connected)
 {
 	emit_property_changed(conn, device->path, DEVICE_INTERFACE,
 				"Connected", DBUS_TYPE_BOOLEAN, &connected);
@@ -612,7 +614,8 @@ void device_set_secmode3_conn(struct btd_device *device, gboolean enable)
 	device->secmode3 = enable;
 }
 
-struct btd_device *device_create(DBusConnection *conn, struct btd_adapter *adapter,
+struct btd_device *device_create(DBusConnection *conn,
+					struct btd_adapter *adapter,
 					const gchar *address)
 {
 	gchar *address_up;
@@ -640,10 +643,13 @@ struct btd_device *device_create(DBusConnection *conn, struct btd_adapter *adapt
 	str2ba(address, &device->bdaddr);
 	device->adapter = adapter;
 
+	device->auth = 0xff;
+
 	return device;
 }
 
-static void device_remove_bonding(struct btd_device *device, DBusConnection *conn)
+static void device_remove_bonding(struct btd_device *device,
+							DBusConnection *conn)
 {
 	char filename[PATH_MAX + 1];
 	char *str, srcaddr[18], dstaddr[18];
@@ -694,8 +700,21 @@ static void device_remove_bonding(struct btd_device *device, DBusConnection *con
 				"Paired", DBUS_TYPE_BOOLEAN, &paired);
 }
 
+static void device_remove_stored(struct btd_device *device,
+					DBusConnection *conn)
+{
+	bdaddr_t src;
+	char addr[18];
+
+	adapter_get_address(device->adapter, &src);
+	ba2str(&device->bdaddr, addr);
+	device_remove_bonding(device, conn);
+	delete_entry(&src, "profiles", addr);
+	delete_entry(&src, "trusts", addr);
+}
+
 void device_remove(struct btd_device *device, DBusConnection *conn,
-						gboolean remove_bonding)
+						gboolean remove_stored)
 {
 	GSList *list;
 	struct btd_device_driver *driver;
@@ -706,8 +725,8 @@ void device_remove(struct btd_device *device, DBusConnection *conn,
 	if (device->bonding)
 		device_cancel_bonding(device, HCI_OE_USER_ENDED_CONNECTION);
 
-	if (!device->temporary && remove_bonding)
-		device_remove_bonding(device, conn);
+	if (!device->temporary && remove_stored)
+		device_remove_stored(device, conn);
 
 	for (list = device->drivers; list; list = list->next) {
 		struct btd_driver_data *driver_data = list->data;
@@ -717,6 +736,7 @@ void device_remove(struct btd_device *device, DBusConnection *conn,
 		g_free(driver_data);
 	}
 
+
 	g_dbus_unregister_interface(conn, path, DEVICE_INTERFACE);
 
 	g_free(path);
@@ -843,13 +863,13 @@ void device_probe_drivers(struct btd_device *device, GSList *profiles)
 
 	for (list = profiles; list; list = list->next) {
 		GSList *l = g_slist_find_custom(device->uuids, list->data,
-							(GCompareFunc) strcasecmp);
+						(GCompareFunc) strcasecmp);
 		if (l)
 			continue;
 
 		device->uuids = g_slist_insert_sorted(device->uuids,
-							g_strdup(list->data),
-							(GCompareFunc) strcasecmp);
+						g_strdup(list->data),
+						(GCompareFunc) strcasecmp);
 	}
 
 	if (device->tmp_records) {
@@ -889,7 +909,8 @@ static void device_remove_drivers(struct btd_device *device, GSList *uuids)
 					(GCompareFunc) strcasecmp))
 				continue;
 
-			debug("UUID %s was removed from device %s", *uuid, dstaddr);
+			debug("UUID %s was removed from device %s",
+							*uuid, dstaddr);
 
 			driver->remove(device);
 			device->drivers = g_slist_remove(device->drivers,
@@ -1094,7 +1115,8 @@ static void update_services(struct browse_req *req, sdp_list_t *recs)
 		store_record(srcaddr, dstaddr, rec);
 
 		/* Copy record */
-		req->records = sdp_list_append(req->records, sdp_copy_record(rec));
+		req->records = sdp_list_append(req->records,
+							sdp_copy_record(rec));
 
 		l = g_slist_find_custom(device->uuids, profile_uuid,
 							(GCompareFunc) strcmp);
@@ -1176,7 +1198,8 @@ static void search_cb(sdp_list_t *recs, int err, gpointer user_data)
 	}
 
 	if (device->tmp_records && req->records) {
-		sdp_list_free(device->tmp_records, (sdp_free_func_t) sdp_record_free);
+		sdp_list_free(device->tmp_records,
+					(sdp_free_func_t) sdp_record_free);
 		device->tmp_records = req->records;
 		req->records = NULL;
 	}
@@ -1227,8 +1250,8 @@ static void browse_cb(sdp_list_t *recs, int err, gpointer user_data)
 	bdaddr_t src;
 	uuid_t uuid;
 
-	/* If we have a valid response and req->search_uuid == 2, then
-	   L2CAP UUID & PNP searching was successful -- we are done */
+	/* If we have a valid response and req->search_uuid == 2, then L2CAP
+	 * UUID & PNP searching was successful -- we are done */
 	if (err < 0 || (req->search_uuid == 2 && req->records))
 		goto done;
 
@@ -1299,12 +1322,14 @@ int device_browse(struct btd_device *device, DBusConnection *conn,
 	device->discov_active = 1;
 
 	if (msg) {
+		const char *sender = dbus_message_get_sender(msg);
+
 		req->msg = dbus_message_ref(msg);
-		device->discov_requestor = g_strdup(dbus_message_get_sender(msg));
+		device->discov_requestor = g_strdup(sender);
 		/* Track the request owner to cancel it
 		 * automatically if the owner exits */
 		device->discov_listener = g_dbus_add_disconnect_watch(conn,
-						dbus_message_get_sender(msg),
+						sender,
 						discover_services_req_exit,
 						device, NULL);
 	}
@@ -1457,7 +1482,7 @@ DBusMessage *new_authentication_return(DBusMessage *msg, uint8_t status)
 	}
 }
 
-static void bonding_request_free(struct bonding_req *bonding)
+static void bonding_request_free(struct bonding_req *bonding, gboolean close)
 {
 	struct btd_device *device;
 
@@ -1476,8 +1501,11 @@ static void bonding_request_free(struct bonding_req *bonding)
 	if (bonding->io_id)
 		g_source_remove(bonding->io_id);
 
-	if (bonding->io)
+	if (bonding->io) {
+		if (close)
+			g_io_channel_close(bonding->io);
 		g_io_channel_unref(bonding->io);
+	}
 
 	device = bonding->device;
 
@@ -1538,7 +1566,6 @@ static struct bonding_req *bonding_request_new(DBusConnection *conn,
 			device->path, name, agent_path);
 
 proceed:
-
 	bonding = g_new0(struct bonding_req, 1);
 
 	bonding->conn = dbus_connection_ref(conn);
@@ -1666,7 +1693,7 @@ failed:
 
 cleanup:
 	device->bonding->io_id = 0;
-	bonding_request_free(device->bonding);
+	bonding_request_free(device->bonding, FALSE);
 
 	return FALSE;
 }
@@ -1678,12 +1705,12 @@ static void create_bond_req_exit(DBusConnection *conn, void *user_data)
 	debug("%s: requestor exited before bonding was completed", device->path);
 
 	if (device->authr)
-		device_cancel_authentication(device);
+		device_cancel_authentication(device, FALSE);
 
 	if (device->bonding) {
 		device->bonding->listener_id = 0;
 		g_io_channel_close(device->bonding->io);
-		bonding_request_free(device->bonding);
+		bonding_request_free(device->bonding, TRUE);
 	}
 }
 
@@ -1816,13 +1843,26 @@ DBusMessage *device_create_bonding(struct btd_device *device,
 	return NULL;
 }
 
+void device_simple_pairing_complete(struct btd_device *device, uint8_t status)
+{
+	struct authentication_req *auth = device->authr;
+
+	if (auth && auth->type == AUTH_TYPE_NOTIFY && auth->agent)
+		agent_cancel(auth->agent);
+}
+
 void device_bonding_complete(struct btd_device *device, uint8_t status)
 {
 	struct bonding_req *bonding = device->bonding;
+	struct authentication_req *auth = device->authr;
+
+	if (auth && auth->type == AUTH_TYPE_NOTIFY && auth->agent)
+		agent_cancel(auth->agent);
 
 	if (status)
 		goto failed;
 
+	device->auth = 0xff;
 	device->temporary = FALSE;
 
 	g_free(device->authr);
@@ -1857,12 +1897,11 @@ void device_bonding_complete(struct btd_device *device, uint8_t status)
 
 	device_set_paired(device, TRUE);
 
-	bonding_request_free(bonding);
+	bonding_request_free(bonding, TRUE);
 
 	return;
 
 failed:
-
 	device_cancel_bonding(device, status);
 }
 
@@ -1890,12 +1929,12 @@ void device_cancel_bonding(struct btd_device *device, uint8_t status)
 	debug("%s: canceling bonding request", device->path);
 
 	if (device->authr)
-		device_cancel_authentication(device);
+		device_cancel_authentication(device, FALSE);
 
 	reply = new_authentication_return(bonding->msg, status);
 	g_dbus_send_message(bonding->conn, reply);
 
-	bonding_request_free(bonding);
+	bonding_request_free(bonding, TRUE);
 }
 
 static void pincode_cb(struct agent *agent, DBusError *err, const char *pincode,
@@ -2032,7 +2071,7 @@ static void cancel_authentication(struct authentication_req *auth)
 	auth->cb = NULL;
 }
 
-void device_cancel_authentication(struct btd_device *device)
+void device_cancel_authentication(struct btd_device *device, gboolean aborted)
 {
 	struct authentication_req *auth = device->authr;
 
@@ -2044,7 +2083,9 @@ void device_cancel_authentication(struct btd_device *device)
 	if (auth->agent)
 		agent_cancel(auth->agent);
 
-	cancel_authentication(auth);
+	if (!aborted)
+		cancel_authentication(auth);
+
 	device->authr = NULL;
 	g_free(auth);
 }
